package java.util.stream;

import java.util.function.BiConsumer;
import java.util.function.BiFunction;
import java.util.function.BinaryOperator;
import java.util.Optional;
import java.util.Comparator;
import java.util.function.Consumer;
import java.util.function.LongConsumer;
import java.util.function.DoubleConsumer;
import java.util.function.IntConsumer;
import java.util.function.ToDoubleFunction;
import java.util.function.ToLongFunction;
import java.util.function.ToIntFunction;
import java.util.function.Function;
import java.util.Objects;
import java.util.function.Predicate;
import java.util.Spliterators;
import java.util.Iterator;
import java.util.function.IntFunction;
import java.util.Spliterator;
import java.util.function.Supplier;

abstract class ReferencePipeline<P_IN, P_OUT> extends AbstractPipeline<P_IN, P_OUT, Stream<P_OUT>> implements Stream<P_OUT>
{
    ReferencePipeline(Supplier<? extends Spliterator<?>> supplier, int n, boolean b) {
        super(supplier, n, b);
    }
    
    ReferencePipeline(Spliterator<?> spliterator, int n, boolean b) {
        super(spliterator, n, b);
    }
    
    ReferencePipeline(AbstractPipeline<?, P_IN, ?> abstractPipeline, int n) {
        super(abstractPipeline, n);
    }
    
    @Override
    final StreamShape getOutputShape() {
        return StreamShape.REFERENCE;
    }
    
    @Override
    final <P_IN> Node<P_OUT> evaluateToNode(PipelineHelper<P_OUT> pipelineHelper, Spliterator<P_IN> spliterator, boolean b, IntFunction<P_OUT[]> intFunction) {
        return Nodes.collect(pipelineHelper, spliterator, b, intFunction);
    }
    
    @Override
    final <P_IN> Spliterator<P_OUT> wrap(PipelineHelper<P_OUT> pipelineHelper, Supplier<Spliterator<P_IN>> supplier, boolean b) {
        return new StreamSpliterators.WrappingSpliterator(pipelineHelper, supplier, b);
    }
    
    @Override
    final Spliterator<P_OUT> lazySpliterator(Supplier<? extends Spliterator<P_OUT>> supplier) {
        return new StreamSpliterators.DelegatingSpliterator(supplier);
    }
    
    @Override
    final void forEachWithCancel(Spliterator<P_OUT> spliterator, Sink<P_OUT> sink) {
        while (!sink.cancellationRequested() && spliterator.tryAdvance(sink)) {}
    }
    
    @Override
    final Node.Builder<P_OUT> makeNodeBuilder(long n, IntFunction<P_OUT[]> intFunction) {
        return Nodes.builder(n, intFunction);
    }
    
    @Override
    public final Iterator<P_OUT> iterator() {
        return Spliterators.iterator(this.spliterator());
    }
    
    @Override
    public Stream<P_OUT> unordered() {
        if (!this.isOrdered()) {
            return this;
        }
        return new ReferencePipeline.StatelessOp<P_OUT, P_OUT>(this, StreamShape.REFERENCE, StreamOpFlag.NOT_ORDERED) {
            @Override
            Sink<P_OUT> opWrapSink(int n, Sink<P_OUT> sink) {
                return sink;
            }
        };
    }
    
    @Override
    public final Stream<P_OUT> filter(final Predicate<? super P_OUT> obj) {
        Objects.requireNonNull(obj);
        return new ReferencePipeline.StatelessOp<P_OUT, P_OUT>(this, StreamShape.REFERENCE, StreamOpFlag.NOT_SIZED) {
            @Override
            Sink<P_OUT> opWrapSink(int n, Sink<P_OUT> sink) {
                return new Sink.ChainedReference<P_OUT, P_OUT>(sink) {
                    @Override
                    public void begin(long n) {
                        this.downstream.begin(-1L);
                    }
                    
                    @Override
                    public void accept(P_OUT p_OUT) {
                        if (obj.test(p_OUT)) {
                            this.downstream.accept(p_OUT);
                        }
                    }
                };
            }
        };
    }
    
    @Override
    public final <R> Stream<R> map(final Function<? super P_OUT, ? extends R> obj) {
        Objects.requireNonNull(obj);
        return new ReferencePipeline.StatelessOp<P_OUT, R>(this, StreamShape.REFERENCE, StreamOpFlag.NOT_SORTED | StreamOpFlag.NOT_DISTINCT) {
            @Override
            Sink<P_OUT> opWrapSink(int n, Sink<R> sink) {
                return new Sink.ChainedReference<P_OUT, R>(sink) {
                    @Override
                    public void accept(P_OUT p_OUT) {
                        this.downstream.accept(obj.apply(p_OUT));
                    }
                };
            }
        };
    }
    
    @Override
    public final IntStream mapToInt(final ToIntFunction<? super P_OUT> obj) {
        Objects.requireNonNull(obj);
        return new IntPipeline.StatelessOp<P_OUT>(this, StreamShape.REFERENCE, StreamOpFlag.NOT_SORTED | StreamOpFlag.NOT_DISTINCT) {
            @Override
            Sink<P_OUT> opWrapSink(int n, Sink<Integer> sink) {
                return new Sink.ChainedReference<P_OUT, Integer>(sink) {
                    @Override
                    public void accept(P_OUT p_OUT) {
                        this.downstream.accept(obj.applyAsInt(p_OUT));
                    }
                };
            }
        };
    }
    
    @Override
    public final LongStream mapToLong(final ToLongFunction<? super P_OUT> obj) {
        Objects.requireNonNull(obj);
        return new LongPipeline.StatelessOp<P_OUT>(this, StreamShape.REFERENCE, StreamOpFlag.NOT_SORTED | StreamOpFlag.NOT_DISTINCT) {
            @Override
            Sink<P_OUT> opWrapSink(int n, Sink<Long> sink) {
                return new Sink.ChainedReference<P_OUT, Long>(sink) {
                    @Override
                    public void accept(P_OUT p_OUT) {
                        this.downstream.accept(obj.applyAsLong(p_OUT));
                    }
                };
            }
        };
    }
    
    @Override
    public final DoubleStream mapToDouble(final ToDoubleFunction<? super P_OUT> obj) {
        Objects.requireNonNull(obj);
        return new DoublePipeline.StatelessOp<P_OUT>(this, StreamShape.REFERENCE, StreamOpFlag.NOT_SORTED | StreamOpFlag.NOT_DISTINCT) {
            @Override
            Sink<P_OUT> opWrapSink(int n, Sink<Double> sink) {
                return new Sink.ChainedReference<P_OUT, Double>(sink) {
                    @Override
                    public void accept(P_OUT p_OUT) {
                        this.downstream.accept(obj.applyAsDouble(p_OUT));
                    }
                };
            }
        };
    }
    
    @Override
    public final <R> Stream<R> flatMap(final Function<? super P_OUT, ? extends Stream<? extends R>> obj) {
        Objects.requireNonNull(obj);
        return new ReferencePipeline.StatelessOp<P_OUT, R>(this, StreamShape.REFERENCE, StreamOpFlag.NOT_SORTED | StreamOpFlag.NOT_DISTINCT | StreamOpFlag.NOT_SIZED) {
            @Override
            Sink<P_OUT> opWrapSink(int n, Sink<R> sink) {
                return new Sink.ChainedReference<P_OUT, R>(sink) {
                    @Override
                    public void begin(long n) {
                        this.downstream.begin(-1L);
                    }
                    
                    @Override
                    public void accept(P_OUT p_OUT) {
                        try (final Stream stream = (Stream)obj.apply(p_OUT)) {
                            if (stream != null) {
                                ((Stream)stream.sequential()).forEach(this.downstream);
                            }
                        }
                    }
                };
            }
        };
    }
    
    @Override
    public final IntStream flatMapToInt(final Function<? super P_OUT, ? extends IntStream> obj) {
        Objects.requireNonNull(obj);
        return new IntPipeline.StatelessOp<P_OUT>(this, StreamShape.REFERENCE, StreamOpFlag.NOT_SORTED | StreamOpFlag.NOT_DISTINCT | StreamOpFlag.NOT_SIZED) {
            @Override
            Sink<P_OUT> opWrapSink(int n, Sink<Integer> sink) {
                return new Sink.ChainedReference<P_OUT, Integer>(sink) {
                    IntConsumer downstreamAsInt = this.downstream::accept;
                    
                    @Override
                    public void begin(long n) {
                        this.downstream.begin(-1L);
                    }
                    
                    @Override
                    public void accept(P_OUT p_OUT) {
                        try (final IntStream intStream = (IntStream)obj.apply(p_OUT)) {
                            if (intStream != null) {
                                intStream.sequential().forEach(this.downstreamAsInt);
                            }
                        }
                    }
                };
            }
        };
    }
    
    @Override
    public final DoubleStream flatMapToDouble(final Function<? super P_OUT, ? extends DoubleStream> obj) {
        Objects.requireNonNull(obj);
        return new DoublePipeline.StatelessOp<P_OUT>(this, StreamShape.REFERENCE, StreamOpFlag.NOT_SORTED | StreamOpFlag.NOT_DISTINCT | StreamOpFlag.NOT_SIZED) {
            @Override
            Sink<P_OUT> opWrapSink(int n, Sink<Double> sink) {
                return new Sink.ChainedReference<P_OUT, Double>(sink) {
                    DoubleConsumer downstreamAsDouble = this.downstream::accept;
                    
                    @Override
                    public void begin(long n) {
                        this.downstream.begin(-1L);
                    }
                    
                    @Override
                    public void accept(P_OUT p_OUT) {
                        try (final DoubleStream doubleStream = (DoubleStream)obj.apply(p_OUT)) {
                            if (doubleStream != null) {
                                doubleStream.sequential().forEach(this.downstreamAsDouble);
                            }
                        }
                    }
                };
            }
        };
    }
    
    @Override
    public final LongStream flatMapToLong(final Function<? super P_OUT, ? extends LongStream> obj) {
        Objects.requireNonNull(obj);
        return new LongPipeline.StatelessOp<P_OUT>(this, StreamShape.REFERENCE, StreamOpFlag.NOT_SORTED | StreamOpFlag.NOT_DISTINCT | StreamOpFlag.NOT_SIZED) {
            @Override
            Sink<P_OUT> opWrapSink(int n, Sink<Long> sink) {
                return new Sink.ChainedReference<P_OUT, Long>(sink) {
                    LongConsumer downstreamAsLong = this.downstream::accept;
                    
                    @Override
                    public void begin(long n) {
                        this.downstream.begin(-1L);
                    }
                    
                    @Override
                    public void accept(P_OUT p_OUT) {
                        try (final LongStream longStream = (LongStream)obj.apply(p_OUT)) {
                            if (longStream != null) {
                                longStream.sequential().forEach(this.downstreamAsLong);
                            }
                        }
                    }
                };
            }
        };
    }
    
    @Override
    public final Stream<P_OUT> peek(final Consumer<? super P_OUT> obj) {
        Objects.requireNonNull(obj);
        return new ReferencePipeline.StatelessOp<P_OUT, P_OUT>(this, StreamShape.REFERENCE, 0) {
            @Override
            Sink<P_OUT> opWrapSink(int n, Sink<P_OUT> sink) {
                return new Sink.ChainedReference<P_OUT, P_OUT>(sink) {
                    @Override
                    public void accept(P_OUT p_OUT) {
                        obj.accept(p_OUT);
                        this.downstream.accept(p_OUT);
                    }
                };
            }
        };
    }
    
    @Override
    public final Stream<P_OUT> distinct() {
        return DistinctOps.makeRef(this);
    }
    
    @Override
    public final Stream<P_OUT> sorted() {
        return SortedOps.makeRef(this);
    }
    
    @Override
    public final Stream<P_OUT> sorted(Comparator<? super P_OUT> comparator) {
        return SortedOps.makeRef(this, comparator);
    }
    
    @Override
    public final Stream<P_OUT> limit(long i) {
        if (i < 0L) {
            throw new IllegalArgumentException(Long.toString(i));
        }
        return SliceOps.makeRef(this, 0L, i);
    }
    
    @Override
    public final Stream<P_OUT> skip(long i) {
        if (i < 0L) {
            throw new IllegalArgumentException(Long.toString(i));
        }
        if (i == 0L) {
            return this;
        }
        return SliceOps.makeRef(this, i, -1L);
    }
    
    @Override
    public void forEach(Consumer<? super P_OUT> consumer) {
        this.evaluate(ForEachOps.makeRef(consumer, false));
    }
    
    @Override
    public void forEachOrdered(Consumer<? super P_OUT> consumer) {
        this.evaluate(ForEachOps.makeRef(consumer, true));
    }
    
    @Override
    public final <A> A[] toArray(IntFunction<A[]> intFunction) {
        return (Object[])Nodes.flatten(this.evaluateToArrayNode(intFunction), intFunction).asArray(intFunction);
    }
    
    @Override
    public final Object[] toArray() {
        return this.toArray(Object[]::new);
    }
    
    @Override
    public final boolean anyMatch(Predicate<? super P_OUT> predicate) {
        return (boolean)this.evaluate(MatchOps.makeRef(predicate, MatchOps.MatchKind.ANY));
    }
    
    @Override
    public final boolean allMatch(Predicate<? super P_OUT> predicate) {
        return (boolean)this.evaluate(MatchOps.makeRef(predicate, MatchOps.MatchKind.ALL));
    }
    
    @Override
    public final boolean noneMatch(Predicate<? super P_OUT> predicate) {
        return (boolean)this.evaluate(MatchOps.makeRef(predicate, MatchOps.MatchKind.NONE));
    }
    
    @Override
    public final Optional<P_OUT> findFirst() {
        return (Optional)this.evaluate(FindOps.makeRef(true));
    }
    
    @Override
    public final Optional<P_OUT> findAny() {
        return (Optional)this.evaluate(FindOps.makeRef(false));
    }
    
    @Override
    public final P_OUT reduce(P_OUT p_OUT, BinaryOperator<P_OUT> binaryOperator) {
        return this.evaluate(ReduceOps.makeRef(p_OUT, binaryOperator, binaryOperator));
    }
    
    @Override
    public final Optional<P_OUT> reduce(BinaryOperator<P_OUT> binaryOperator) {
        return (Optional)this.evaluate(ReduceOps.makeRef(binaryOperator));
    }
    
    @Override
    public final <R> R reduce(R r, BiFunction<R, ? super P_OUT, R> biFunction, BinaryOperator<R> binaryOperator) {
        return this.evaluate(ReduceOps.makeRef(r, biFunction, binaryOperator));
    }
    
    @Override
    public final <R, A> R collect(Collector<? super P_OUT, A, R> collector2) {
        Object o;
        if (this.isParallel() && collector2.characteristics().contains(Collector.Characteristics.CONCURRENT) && (!this.isOrdered() || collector2.characteristics().contains(Collector.Characteristics.UNORDERED))) {
            o = collector2.supplier().get();
            this.forEach(o2/* java.lang.Object, */ -> {
                collector2.accumulator();
                biConsumer.accept(var_2_3B, o2);
                return;
            });
        }
        else {
            o = this.evaluate(ReduceOps.makeRef(collector2));
        }
        return collector2.characteristics().contains(Collector.Characteristics.IDENTITY_FINISH) ? o : collector2.finisher().apply(o);
    }
    
    @Override
    public final <R> R collect(Supplier<R> supplier, BiConsumer<R, ? super P_OUT> biConsumer, BiConsumer<R, R> biConsumer2) {
        return this.evaluate(ReduceOps.makeRef(supplier, biConsumer, biConsumer2));
    }
    
    @Override
    public final Optional<P_OUT> max(Comparator<? super P_OUT> comparator) {
        return this.reduce(BinaryOperator.maxBy(comparator));
    }
    
    @Override
    public final Optional<P_OUT> min(Comparator<? super P_OUT> comparator) {
        return this.reduce(BinaryOperator.minBy(comparator));
    }
    
    @Override
    public final long count() {
        return this.mapToLong(p0/* java.lang.Object, */ -> 1L).sum();
    }
    
    static class Head<E_IN, E_OUT> extends ReferencePipeline<E_IN, E_OUT>
    {
        Head(Supplier<? extends Spliterator<?>> supplier, int n, boolean b) {
            super(supplier, n, b);
        }
        
        Head(Spliterator<?> spliterator, int n, boolean b) {
            super(spliterator, n, b);
        }
        
        @Override
        final boolean opIsStateful() {
            throw new UnsupportedOperationException();
        }
        
        @Override
        final Sink<E_IN> opWrapSink(int n, Sink<E_OUT> sink) {
            throw new UnsupportedOperationException();
        }
        
        @Override
        public void forEach(Consumer<? super E_OUT> action) {
            if (!this.isParallel()) {
                this.sourceStageSpliterator().forEachRemaining(action);
            }
            else {
                super.forEach(action);
            }
        }
        
        @Override
        public void forEachOrdered(Consumer<? super E_OUT> action) {
            if (!this.isParallel()) {
                this.sourceStageSpliterator().forEachRemaining(action);
            }
            else {
                super.forEachOrdered(action);
            }
        }
    }
    
    abstract static class StatelessOp<E_IN, E_OUT> extends ReferencePipeline<E_IN, E_OUT>
    {
        StatelessOp(AbstractPipeline<?, E_IN, ?> abstractPipeline, StreamShape streamShape, int n) {
            super(abstractPipeline, n);
            assert abstractPipeline.getOutputShape() == streamShape;
            return;
        }
        
        @Override
        final boolean opIsStateful() {
            return false;
        }
    }
    
    abstract static class StatefulOp<E_IN, E_OUT> extends ReferencePipeline<E_IN, E_OUT>
    {
        StatefulOp(AbstractPipeline<?, E_IN, ?> abstractPipeline, StreamShape streamShape, int n) {
            super(abstractPipeline, n);
            assert abstractPipeline.getOutputShape() == streamShape;
            return;
        }
        
        @Override
        final boolean opIsStateful() {
            return true;
        }
        
        @Override
        abstract <P_IN> Node<E_OUT> opEvaluateParallel(PipelineHelper<E_OUT> p0, Spliterator<P_IN> p1, IntFunction<E_OUT[]> p2);
    }
}
